VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "DirectSoundOut2"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit


' DirectSound PCM Output Stream
'
' Arne Elster 2009


Public Enum DSO_STATUS
    DSO_PLAYING
    DSO_PAUSING
    DSO_STOPPED
End Enum

Private WithEvents m_clsTmr     As Timer
Attribute m_clsTmr.VB_VarHelpID = -1
Private m_udtFormat             As DSBUFFERDESC
Private m_DXBuffer              As DirectSoundSecondaryBuffer8
Private m_clsData               As IDataCallback

Private m_sngBuffer()           As Single
Private m_lngBufferSamples      As Long

Private m_udeStatus             As DSO_STATUS
Private m_blnEOS                As Boolean
Private m_blnReady              As Boolean

Private m_lngPositionMS         As Long
Private m_lngDataStart          As Long
Private m_lngDataLength         As Long
Private m_lngEOSPosition        As Long
Private m_lngEOSCycle           As Long

Private m_lngCycleCount         As Long
Private m_lngLastPosition       As Long

Private Const TIMER_INTERVAL    As Long = 25
Private Const DATA_MIN_LEN_MS   As Long = 500
Private Const STREAM_EOS_CUTOFF As Long = 10


Private Sub Class_Initialize()
    Set m_clsTmr = TimerManager.CreateNewTimer(TIMER_INTERVAL, False)
    m_udeStatus = DSO_STOPPED
End Sub


Public Sub Terminate()
    m_clsTmr.enabled = False
    Set m_DXBuffer = Nothing
    m_blnReady = False
End Sub

Public Function Initialize( _
    ByVal lngSamplerate As Long, _
    ByVal lngChannels As Long, _
    clsDataCallback As IDataCallback, _
    Optional ByVal lngBufferMS As Long = 2000 _
) As Boolean

    Terminate

    m_udtFormat.fxFormat = CreateWFX(lngSamplerate, lngChannels, 16)
    m_udtFormat.lBufferBytes = MSToBytes(lngBufferMS)
    m_udtFormat.lFlags = DSBCAPS_CTRLPAN Or _
                         DSBCAPS_CTRLVOLUME Or _
                         DSBCAPS_GETCURRENTPOSITION2 Or _
                         DSBCAPS_GLOBALFOCUS

    Set m_DXBuffer = DirectSound.CreateSoundBuffer(m_udtFormat)
    m_lngBufferSamples = MSToSamples(lngBufferMS)
    ReDim m_sngBuffer(m_lngBufferSamples - 1) As Single
    
    Set m_clsData = clsDataCallback
    m_lngPositionMS = 0
    m_lngEOSPosition = -1
    m_lngEOSCycle = -1
    m_lngLastPosition = 0
    m_lngCycleCount = 0
    m_udeStatus = DSO_STOPPED
    m_blnEOS = False
    m_blnReady = True
    
    Initialize = True
End Function


Public Property Get Volume() As Long
    If Not m_blnReady Then Exit Property
    Volume = m_DXBuffer.GetVolume
End Property

Public Property Let Volume(ByVal lngVal As Long)
    If Not m_blnReady Then Exit Property
    If lngVal > 0 Then lngVal = 0
    If lngVal < -10000 Then lngVal = -10000
    m_DXBuffer.SetVolume lngVal
End Property


Public Property Get Panning() As Long
    If Not m_blnReady Then Exit Property
    Panning = m_DXBuffer.GetPan
End Property

Public Property Let Panning(ByVal lngVal As Long)
    If Not m_blnReady Then Exit Property
    m_DXBuffer.SetPan lngVal
End Property


Public Property Get Status() As DSO_STATUS
    Status = m_udeStatus
End Property


Public Property Get Position() As Long
    Dim udtCursor As DSCURSORS
    Dim lngAdd As Long
    
    m_DXBuffer.GetCurrentPosition udtCursor
    
    If udtCursor.lPlay < m_lngDataStart Then
        lngAdd = BytesToMS(m_udtFormat.lBufferBytes - m_lngDataStart + udtCursor.lPlay)
    Else
        lngAdd = BytesToMS(udtCursor.lPlay - m_lngDataStart)
    End If
    
    Position = m_lngPositionMS + lngAdd
End Property

Public Property Let Position(ByVal lngPos As Long)
    ' Nicht ganz richtig gelst.
    ' Falls Position gendert wird whrend Buffer spielt,
    ' was in Monoton Revived nicht der Fall ist,
    ' wird die Entfernung vom letzten Zeitmesspunkt
    ' auer Acht gelassen
    m_lngPositionMS = lngPos
End Property


Public Sub SetStatus(ByVal stat As DSO_STATUS)
    If Not m_blnReady Then Exit Sub
    
    Select Case stat
        Case DSO_PAUSING
            If m_udeStatus = DSO_PAUSING Then
                m_DXBuffer.Play DSBPLAY_LOOPING
                m_udeStatus = DSO_PLAYING
                m_clsTmr.enabled = True
            Else
                m_DXBuffer.Stop
                m_udeStatus = DSO_PAUSING
                m_clsTmr.enabled = False
            End If
        Case DSO_STOPPED
            m_DXBuffer.Stop
            m_udeStatus = DSO_STOPPED
            m_clsTmr.enabled = False
        Case DSO_PLAYING
            If m_udeStatus = DSO_PAUSING Then
                m_DXBuffer.Play DSBPLAY_LOOPING
                m_clsTmr.enabled = True
            ElseIf m_udeStatus = DSO_STOPPED Then
                m_lngDataStart = 0
                m_lngDataLength = 0
                FillBuffer
                m_DXBuffer.SetCurrentPosition 0
                m_DXBuffer.Play DSBPLAY_LOOPING
                m_clsTmr.enabled = True
            End If
            m_udeStatus = DSO_PLAYING
    End Select
End Sub


Public Sub ResetBuffer()
    m_lngDataStart = 0
    m_lngDataLength = 0
    
    If m_udeStatus = DSO_PLAYING Then
        m_DXBuffer.Stop
        FillBuffer
        m_DXBuffer.Play DSBPLAY_LOOPING
    End If
End Sub


Private Sub m_clsTmr_Timer()
    Dim udtCursor As DSCURSORS
    Dim lngDiff As Long
    
    If m_udeStatus = DSO_PLAYING Then
        m_DXBuffer.GetCurrentPosition udtCursor
        
        If udtCursor.lPlay < m_lngLastPosition Then
            m_lngCycleCount = m_lngCycleCount + 1
        End If
        m_lngLastPosition = udtCursor.lPlay
        
        ' seit letzter Messung gespielte Datenmenge berechnen
        If udtCursor.lPlay < m_lngDataStart Then
            lngDiff = m_udtFormat.lBufferBytes - m_lngDataStart + udtCursor.lPlay
        Else
            lngDiff = udtCursor.lPlay - m_lngDataStart
        End If

        If m_blnEOS Then
            ' Wir erwarten das Ende des Streams
            If m_lngCycleCount >= m_lngEOSCycle Then
                If BytesToMS(udtCursor.lPlay - m_lngEOSPosition) >= 0 Then
                    m_DXBuffer.Stop
                    m_udeStatus = DSO_STOPPED
                    m_clsData.EndOfStream
                End If
            End If
        Else
            ' wenn nur noch X ms Audiodaten im Puffer bereitstehen,
            ' Puffer wieder auffllen
            If BytesToMS(m_lngDataLength - lngDiff) < DATA_MIN_LEN_MS Then
                If m_lngDataLength - lngDiff < 0 Then
                    m_DXBuffer.Stop
                    m_lngDataStart = (m_lngDataStart + m_lngDataLength) Mod m_udtFormat.lBufferBytes
                    m_DXBuffer.SetCurrentPosition m_lngDataStart
                    m_lngPositionMS = m_lngPositionMS + BytesToMS(m_lngDataLength)
                    m_lngDataLength = 0
                    FillBuffer
                    m_DXBuffer.Play DSBPLAY_LOOPING
                Else
                    m_lngPositionMS = m_lngPositionMS + BytesToMS(lngDiff)
                    m_lngDataStart = udtCursor.lPlay
                    m_lngDataLength = m_lngDataLength - lngDiff
                    FillBuffer
                End If
            End If
        End If
    End If
End Sub


Private Sub FillBuffer()
    Dim lngSamplesNeeded As Long
    Dim lngSamplesGot As Long
    Dim lngNewSamples As Long
    Dim lngOffset As Long
    Dim i As Long
    
    ' Offset im DS Buffer, an das neue Samples geschrieben werden
    lngOffset = BytesToSamples(m_lngDataStart + m_lngDataLength) Mod m_lngBufferSamples
    lngSamplesNeeded = MSToSamples(DATA_MIN_LEN_MS * 2)
    
    ' Wrap-Around im DirectSound Buffer beachten
    If lngOffset + lngSamplesNeeded > m_lngBufferSamples Then
        m_clsData.NeedData m_sngBuffer, lngOffset, _
                           m_lngBufferSamples - lngOffset, _
                           lngSamplesGot
        
        lngNewSamples = lngSamplesGot
        
        lngSamplesNeeded = lngSamplesNeeded - (m_lngBufferSamples - lngOffset)
        m_clsData.NeedData m_sngBuffer, 0, lngSamplesNeeded, lngSamplesGot
        
        lngNewSamples = lngNewSamples + lngSamplesGot
    Else
        m_clsData.NeedData m_sngBuffer, lngOffset, lngSamplesNeeded, lngSamplesGot
        lngNewSamples = lngSamplesGot
    End If
    
    If lngNewSamples > 0 Then
        WriteDataToBuffer lngOffset, lngNewSamples
    End If
    
    m_lngDataLength = m_lngDataLength + SamplesToBytes(lngNewSamples)
    
    ' EOS msste hier nur beim zweiten NeedData beachtet werden,
    ' denn wenn der Stream schon leer ist, kann an Offset 0
    ' nichts mehr geschrieben werden
    If lngSamplesGot < lngSamplesNeeded Then
        m_blnEOS = True
        m_lngEOSPosition = (m_lngDataStart + m_lngDataLength) Mod m_udtFormat.lBufferBytes
        
        If m_lngDataStart + m_lngDataLength > m_udtFormat.lBufferBytes Then
            ' EOS tritt erst bei nchstem Bufferdurchgang auf
            ' daher nicht zu frh auslsen
            m_lngEOSCycle = m_lngCycleCount + 1
        Else
            m_lngEOSCycle = m_lngCycleCount
        End If
        
        ' Timer kann die Wiedergabe wohl in den seltensten Fllen
        ' genau bei EOS abbrechen, deswegen Buffer nach hinten
        ' mit Stille fllen, um nervige Klicks und Kratzer
        ' zu vermeiden
        PadBufferWithSilence BytesToSamples(m_lngDataStart + m_lngDataLength) Mod m_lngBufferSamples
    End If
End Sub


' Samples von Float in 16 Bit Int umrechnen
' und dann in den DS Buffer schreiben
Private Sub WriteDataToBuffer(ByVal lngOffset As Long, ByVal lngSamples As Long)
    Dim intSamples() As Integer
    ReDim intSamples(lngSamples - 1) As Integer
    Dim i As Long
    Dim sngSample As Single
    
    For i = lngOffset To lngOffset + lngSamples - 1
        sngSample = m_sngBuffer(i Mod m_lngBufferSamples)
        
        If sngSample > 1 Then
            sngSample = 1
        ElseIf sngSample < -1 Then
            sngSample = -1
        End If
        
        intSamples(i - lngOffset) = sngSample * 32767
    Next
    
    m_DXBuffer.WriteBuffer SamplesToBytes(lngOffset), _
                           SamplesToBytes(lngSamples), _
                           intSamples(0), DSBLOCK_DEFAULT
End Sub


' 2 Timer Intervalle Stille in den Buffer schreiben
' (Timer erkennt ob alle Samples gespielt wurden,
'  DS kann also ber letztes Sample weiterspielen
'  wegen Auflsung von Timer)
Private Sub PadBufferWithSilence(ByVal lngOffset As Long)
    Dim lngSamples As Long
    Dim intSilence() As Integer
    
    lngSamples = MSToSamples(m_clsTmr.interval * 2)
    ReDim intSilence(lngSamples - 1) As Integer
    
    m_DXBuffer.WriteBuffer SamplesToBytes(lngOffset), _
                           SamplesToBytes(lngSamples), _
                           intSilence(0), DSBLOCK_DEFAULT
End Sub


Public Function CaptureSamples( _
    sngSamples() As Single, _
    ByVal lngOffset As Long, _
    ByVal lngSamples As Long, _
    ByRef lngWritten As Long _
) As Boolean

    If Not m_blnReady Then Exit Function

    Dim udtPos      As DSCURSORS
    Dim lngCapStart As Long

    m_DXBuffer.GetCurrentPosition udtPos
    lngCapStart = BytesToSamples(udtPos.lPlay)
    lngWritten = 0
    If lngSamples > m_lngBufferSamples Then Exit Function

    If lngCapStart + lngSamples > m_lngBufferSamples Then
        Dim lngDiff As Long
        lngDiff = m_lngBufferSamples - lngCapStart
        DXCopyMemory sngSamples(lngOffset), m_sngBuffer(lngCapStart), 4 * lngDiff
        lngSamples = lngSamples - lngDiff
        lngWritten = lngWritten + lngDiff
        lngOffset = lngOffset + lngWritten
        lngCapStart = (lngCapStart + lngWritten) Mod m_lngBufferSamples
    End If
    
    DXCopyMemory sngSamples(lngOffset), m_sngBuffer(lngCapStart), 4 * lngSamples
    lngWritten = lngWritten + lngSamples

    CaptureSamples = True
End Function


Private Function BytesToSamples(ByVal bytes As Long) As Long
    BytesToSamples = bytes / (m_udtFormat.fxFormat.nBitsPerSample \ 8)
End Function

Private Function SamplesToBytes(ByVal smp As Long) As Long
    SamplesToBytes = smp * (m_udtFormat.fxFormat.nBitsPerSample \ 8)
End Function

Private Function BytesToMS(ByVal bytes As Long) As Long
    BytesToMS = bytes / m_udtFormat.fxFormat.lAvgBytesPerSec * 1000
End Function

Private Function MSToBytes(ByVal ms As Long) As Long
    MSToBytes = ms / 1000 * m_udtFormat.fxFormat.lAvgBytesPerSec
End Function

Private Function SamplesToMS(ByVal smp As Long) As Long
    SamplesToMS = BytesToMS(smp * m_udtFormat.fxFormat.nBitsPerSample \ 8)
End Function

Private Function MSToSamples(ByVal ms As Long) As Long
    MSToSamples = MSToBytes(ms) \ (m_udtFormat.fxFormat.nBitsPerSample \ 8)
End Function
