VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "Player"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit


' Verbindung zwischen Decoder und Ausgabe
'
' Arne Elster 2009


Implements IDataCallback

Private m_clsDSOut      As DirectSoundOut2
Private m_clsStream     As IPluginStream

Private m_blnEOS        As Boolean

Private m_lngMuteVol    As Long
Private m_blnMute       As Boolean

Public Event StatusChanged(ByVal Status As DSO_STATUS)
Public Event EndOfStream()

Public Property Get Stream() As IPluginStream
    Set Stream = m_clsStream
End Property

Public Property Get Mute() As Boolean
    Mute = m_blnMute
End Property

Public Property Let Mute(ByVal value As Boolean)
    If value Then
        If Not m_blnMute Then m_lngMuteVol = m_clsDSOut.Volume
        m_clsDSOut.Volume = -10000
    Else
        m_clsDSOut.Volume = m_lngMuteVol
    End If
    m_blnMute = value
End Property

Private Sub Class_Initialize()
    Set m_clsDSOut = New DirectSoundOut2
End Sub

Private Sub Class_Terminate()
    Set m_clsDSOut = Nothing
End Sub

Public Sub Play(Optional ByVal blnFade As Boolean = False)
    Dim lngVol As Long
        
    If blnFade Then
        lngVol = m_clsDSOut.Volume
        m_clsDSOut.Volume = -10000
    End If
    
    m_clsDSOut.SetStatus DSO_PLAYING
    
    If blnFade Then
        RampVolume -10000, lngVol
        m_clsDSOut.Volume = lngVol
    End If
    
    RaiseEvent StatusChanged(m_clsDSOut.Status)
End Sub

Public Sub StopPlay(Optional ByVal blnFade As Boolean = False)
    Dim lngVol As Long
    
    If Not m_clsStream Is Nothing Then
        If blnFade Then
            lngVol = m_clsDSOut.Volume
            RampVolume lngVol, -10000
        End If
        
        m_clsDSOut.SetStatus DSO_STOPPED
        
        If blnFade Then m_clsDSOut.Volume = lngVol
        
        m_blnEOS = False
        m_clsStream.StreamSeek 0
        RaiseEvent StatusChanged(m_clsDSOut.Status)
    End If
End Sub

Public Sub PausePlay(Optional ByVal blnFade As Boolean = False)
    Dim lngVol As Long
    
    If m_clsDSOut.Status = DSO_PLAYING Then
        If blnFade Then
            lngVol = m_clsDSOut.Volume
            RampVolume lngVol, -10000
        End If
        
        m_clsDSOut.SetStatus DSO_PAUSING
        
        If blnFade Then m_clsDSOut.Volume = lngVol
    ElseIf m_clsDSOut.Status = DSO_PAUSING Then
        If blnFade Then
            lngVol = m_clsDSOut.Volume
            m_clsDSOut.Volume = -10000
        End If
        
        m_clsDSOut.SetStatus DSO_PLAYING
        
        If blnFade Then
            RampVolume -10000, lngVol
            m_clsDSOut.Volume = lngVol
        End If
    End If
    
    RaiseEvent StatusChanged(m_clsDSOut.Status)
End Sub

Private Sub RampVolume(ByVal lngVolFrom As Long, ByVal lngVolTo As Long)
    Dim blnDown As Boolean
    Dim lngVolS As Long
    Dim i As Long
    
    blnDown = lngVolFrom > lngVolTo
    lngVolS = (lngVolTo - lngVolFrom) \ 2
    If Not blnDown Then lngVolFrom = lngVolFrom + lngVolTo + 10000
    
    For i = 1 To 99 Step 2
        ' lineare Lautstrkeskala auf logarithmische von DS bertragen
        If blnDown Then
            m_clsDSOut.Volume = lngVolFrom - lngVolS * (Log(1 - i / 100) / LOG10)
        Else
            m_clsDSOut.Volume = lngVolFrom + lngVolS * (Log(i / 100) / LOG10)
        End If
        Wait 1
        
        ' DoEvents bentigt damit whrend der Rampe
        ' der DS Buffer weiter gefllt werden kann
        If i Mod 10 = 0 Then DoEvents
    Next
End Sub

Public Property Get Status() As DSO_STATUS
    Status = m_clsDSOut.Status
End Property

Public Property Get Pan() As Long
    Pan = m_clsDSOut.Panning
End Property

Public Property Let Pan(ByVal lngPan As Long)
    m_clsDSOut.Panning = lngPan
End Property

Public Property Get Volume() As Long
    If m_blnMute Then
        Volume = m_lngMuteVol
    Else
        Volume = m_clsDSOut.Volume
    End If
End Property

Public Property Let Volume(ByVal lngVol As Long)
    If m_blnMute Then
        m_lngMuteVol = lngVol
    Else
        m_clsDSOut.Volume = lngVol
    End If
End Property

Public Property Get PositionMS() As Long
    PositionMS = m_clsDSOut.Position
End Property

Public Property Get DurationMS() As Long
    DurationMS = m_clsStream.Duration
End Property

Public Function CaptureSamples( _
    sngSamples() As Single, _
    ByVal lngOffset As Long, _
    ByVal lngSamples As Long, _
    ByRef lngWritten As Long _
) As Boolean

    CaptureSamples = m_clsDSOut.CaptureSamples(sngSamples, _
                                               lngOffset, _
                                               lngSamples, _
                                               lngWritten)
End Function

Public Function CloseFile() As STREAM_RESULT
    If Not m_clsStream Is Nothing Then
        CloseFile = m_clsStream.StreamClose()
        Set m_clsStream = Nothing
    End If
End Function

Public Function OpenFile(ByVal strFile As String) As STREAM_RESULT
    Dim strExtension As String
    
    strExtension = GetExtension(strFile)
    Set m_clsStream = Factory.GetNewStreamForExt(strExtension)
    
    If m_clsStream Is Nothing Then
        OpenFile = RESULT_INVALID_FORMAT
    Else
        If m_clsStream.StreamOpen(strFile) = RESULT_SUCCESS Then
            If m_clsDSOut.Initialize(m_clsStream.samplerate, _
                                     m_clsStream.Channels, _
                                     Me) Then
                
                DSP.FormatChanged m_clsStream.samplerate, m_clsStream.Channels
                DSP.NextTrack
                
                OpenFile = RESULT_SUCCESS
            Else
                OpenFile = RESULT_INVALID_FORMAT
            End If
        Else
            OpenFile = RESULT_CANT_OPEN
        End If
    End If
End Function

Public Function SeekFile( _
    ByVal ms As Long, _
    Optional ByVal blnFade As Boolean = False _
) As STREAM_RESULT

    Dim lngVol As Long
    
    If Not m_clsStream Is Nothing Then
        If blnFade Then
            lngVol = m_clsDSOut.Volume
            RampVolume lngVol, -10000
        End If
        
        SeekFile = m_clsStream.StreamSeek(ms)
        If SeekFile = RESULT_SUCCESS Then
            m_clsDSOut.ResetBuffer
            m_clsDSOut.Position = m_clsStream.Position
            
            If blnFade Then
                RampVolume -10000, lngVol
                m_clsDSOut.Volume = lngVol
            End If
        End If
    Else
        SeekFile = RESULT_UNKNOWN_ERROR
    End If
End Function

Private Sub IDataCallback_EndOfStream()
    RaiseEvent EndOfStream
End Sub

Private Sub IDataCallback_NeedData( _
    sngSamples() As Single, _
    ByVal lngOffset As Long, _
    ByVal lngSamples As Long, _
    ByRef lngSamplesWritten As Long _
)

    If Not m_clsStream Is Nothing Then
        If m_clsStream.StreamRead(sngSamples, _
                                  lngOffset, _
                                  lngSamples, _
                                  lngSamplesWritten) <> RESULT_SUCCESS Then
            m_blnEOS = True
        End If
        DSP.Process sngSamples, lngOffset, lngSamplesWritten, lngSamplesWritten
    End If
End Sub
