Attribute VB_Name = "DirectXInit"
Option Explicit

' Initialisiert DirectX fr Monoton und
' erstellt den Primary Direct Sound Buffer (Mix Buffer)

Public DirectX              As DirectX8
Public DirectSound          As DirectSound8
Public DirectSoundPrimBuf   As DirectSoundPrimaryBuffer8
Public DirectSoundEnum      As DirectSoundEnum8

Public Function InitializeDirectX() As Boolean
    On Error GoTo ErrorHandler

    If DirectX Is Nothing Then
        Set DirectX = New DirectX8
        Set DirectSoundEnum = DirectX.GetDSEnum
    End If

    InitializeDirectX = True

ErrorHandler:
End Function

Public Property Get DS_DeviceCount() As Long
    If Not InitializeDirectX Then
        Exit Property
    End If

    DS_DeviceCount = DirectSoundEnum.GetCount
End Property

Public Property Get DS_DeviceGUID(ByVal index As Long) As String
    DS_DeviceGUID = DirectSoundEnum.GetGuid(index)
End Property

Public Property Get DS_DeviceName(ByVal index As Long) As String
    DS_DeviceName = DirectSoundEnum.GetName(index)
End Property

Public Property Get DS_DeviceDescription(ByVal index As Long) As String
    DS_DeviceDescription = DirectSoundEnum.GetDescription(index)
End Property

Public Property Get DirectSoundReady() As Boolean
    DirectSoundReady = Not DirectSoundPrimBuf Is Nothing
End Property

Public Function InitDirectSound( _
    ByVal samplerate As Long, _
    ByVal Channels As Integer, _
    ByVal BitsPerSample As Integer, _
    Optional ByVal device_guid As String _
) As Boolean

    On Error GoTo ErrorHandler

    Set DirectSoundPrimBuf = Nothing
    Set DirectSound = Nothing

    Set DirectSound = DirectX.DirectSoundCreate(device_guid)

    DirectSound.SetCooperativeLevel MonoMain.hWnd, _
                                    DSSCL_PRIORITY

    Set DirectSoundPrimBuf = CreatePrimBuf(samplerate, _
                                           Channels, _
                                           BitsPerSample)

    InitDirectSound = True

ErrorHandler:
End Function

Private Function CreatePrimBuf( _
    ByVal samplerate As Long, _
    ByVal Channels As Integer, _
    ByVal BitsPerSample As Long _
) As DirectSoundPrimaryBuffer8

    Dim udtFormat       As DSBUFFERDESC
    Dim DXPrimaryBuffer As DirectSoundPrimaryBuffer8

    With udtFormat
        .fxFormat = CreateWFX(samplerate, Channels, BitsPerSample)
        .lBufferBytes = 0
        .lFlags = DSBCAPS_PRIMARYBUFFER
    End With

    Set DXPrimaryBuffer = DirectSound.CreatePrimarySoundBuffer(udtFormat)
    Set CreatePrimBuf = DXPrimaryBuffer
End Function

Public Function CreateWFX( _
    ByVal sr As Long, _
    ByVal ch As Long, _
    ByVal bps As Long _
) As WAVEFORMATEX

    With CreateWFX
        .nSize = LenB(CreateWFX)
        .nFormatTag = WAVE_FORMAT_PCM
        .lSamplesPerSec = sr
        .nChannels = ch
        .nBitsPerSample = bps
        .nBlockAlign = .nChannels * .nBitsPerSample \ 8
        .lAvgBytesPerSec = .nBlockAlign * .lSamplesPerSec
        .lExtra = 0
    End With
End Function
