VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "ParserGlobal"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit

#Const DEBUG_INTERRPUT = False

Implements IStatement
Implements ParserCodeBlock

Private m_colGlobal As Collection
Private m_colVars   As Collection
Private m_colFncs   As Collection

Public Property Get VarBytes(ByVal ude As DATA_TYPE) As Long
    Select Case ude And (Not TYPE_ARRAY)
        Case TYPE_BOOL:     VarBytes = 4
        Case TYPE_FLOAT:    VarBytes = 4
        Case TYPE_BYTE:     VarBytes = 1
        Case TYPE_INT:      VarBytes = 4
        Case TYPE_SHORT:    VarBytes = 2
        Case TYPE_NONE:     VarBytes = 0
        Case Else:          Err.Raise 1234, , "Unbekannter Typ"
    End Select
End Property

Public Sub AddStatement(ByVal cls As IStatement)
    m_colGlobal.Add cls
End Sub

Public Sub AddFunction(ByVal cls As ParserFunction)
    On Error Resume Next
    m_colFncs.Add cls, cls.Name
End Sub

Public Property Get Statements() As Collection
    Set Statements = m_colGlobal
End Property

Public Property Get Functions() As Collection
    Set Functions = m_colFncs
End Property

Private Sub Class_Initialize()
    Set m_colVars = New Collection
    Set m_colFncs = New Collection
    Set m_colGlobal = New Collection
End Sub

Private Sub IStatement_DeclareVars(ByVal ctx As TACContext)
    Dim clsVar  As ParserStatementDim
    
    For Each clsVar In m_colGlobal
        ToIStatement(clsVar).DeclareVars ctx
    Next
End Sub

Private Sub IStatement_DestroyVars(ByVal ctx As TACContext)
    Dim clsVar  As ParserStatementDim
    
    For Each clsVar In m_colGlobal
        ToIStatement(clsVar).DestroyVars ctx
    Next
End Sub

Private Function IStatement_GenerateTAC(ByVal ctx As TACContext) As TACVar
    Dim clsSt  As IStatement
    
    '##################
    
    ctx.AddCode CreateTAC(TAC_CMD_ENTER, , "global_init")
    
    IStatement_DeclareVars ctx
    
    For Each clsSt In Statements
        clsSt.GenerateTAC ctx
    Next
    
    ctx.AddCode CreateTAC(TAC_CMD_LEAVE, , "global_init")
    
    '##################
    
    ctx.AddCode CreateTAC(TAC_CMD_ENTER, , "global_free")
    IStatement_DestroyVars ctx
    ctx.AddCode CreateTAC(TAC_CMD_LEAVE, , "global_free")
    
    '##################
    
    ctx.AddCode CreateTAC(TAC_CMD_ENTER, , "main")
    
    #If DEBUG_INTERRPUT Then
        ctx.AddCode CreateTAC(TAC_CMD_DEBUG)
    #End If
    
    Dim clsRet As TACVar
    Set clsRet = CreateTempTACVar(ctx, TYPE_INT)
    
    ctx.AddCode CreateTAC(TAC_CMD_CALL, , "global_init")
    ctx.AddCode CreateTAC(TAC_CMD_CALL, , "fnc_main")
    ctx.AddCode CreateTAC(TAC_CMD_RETRIEVE, , , clsRet)
    ctx.AddCode CreateTAC(TAC_CMD_CALL, , "global_free")
    ctx.AddCode CreateTAC(TAC_CMD_RETURN, , , clsRet)
    ctx.AddCode CreateTAC(TAC_CMD_EXITLABEL, , EXIT_PREFIX & "main")
    ctx.AddCode CreateTAC(TAC_CMD_LEAVE, , "main")
    
    For Each clsSt In Functions
        clsSt.GenerateTAC ctx
    Next
End Function

Private Sub ParserCodeBlock_AddStatement(ByVal cls As IStatement)
    m_colGlobal.Add cls
End Sub

Private Sub ParserCodeBlock_AddVariables(ByVal cls As ParserStatementDim)
    m_colVars.Add cls
End Sub

Private Property Get ParserCodeBlock_Statements() As Collection
    Set ParserCodeBlock_Statements = m_colGlobal
End Property

Private Property Get ParserCodeBlock_Variables() As Collection
    Set ParserCodeBlock_Variables = m_colVars
End Property
