VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "IIRFilter"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit

' IIR Filter Implementation
' nach RBJs EQ Cookbook
'
' Arne Elster 2007/08

Private Const Pi        As Single = 3.14159265358979
Private Const LN2       As Single = 0.693147180559945

Public Enum IIR_TYPE
    IIR_LOW_PASS
    IIR_HIGH_PASS
    IIR_BAND_PASS
    IIR_PEAK_EQ
End Enum

Private fa0             As Single
Private fa1             As Single
Private fa2             As Single
Private fa3             As Single
Private fa4             As Single

Private fX1             As Single
Private fX2             As Single
Private fY1             As Single
Private fY2             As Single

Private lngSR           As Long
Private udeType         As IIR_TYPE
Private sngDB           As Single
Private sngFreq         As Single
Private sngBW           As Single

Public Property Get Bandwidth() As Single
    Bandwidth = sngBW
End Property

Public Property Let Bandwidth(ByVal sngVal As Single)
    sngBW = sngVal
    UpdateSettings
End Property

Public Property Get IIRType() As IIR_TYPE
    IIRType = udeType
End Property

Public Property Get GainDB() As Single
    GainDB = sngDB
End Property

Public Property Let GainDB(ByVal sngVal As Single)
    sngDB = sngVal
    UpdateSettings
End Property

Public Property Get Frequency() As Single
    Frequency = sngFreq
End Property

Public Property Let Frequency(ByVal sngVal As Single)
    sngFreq = sngVal
    UpdateSettings
End Property

Public Property Get samplerate() As Long
    samplerate = lngSR
End Property

Public Property Let samplerate(ByVal lngVal As Long)
    lngSR = lngVal
    UpdateSettings
End Property

Private Sub UpdateSettings()
    If lngSR > 0 Then
        CreateBiquadIIR udeType, sngDB, sngFreq, lngSR, sngBW
    End If
End Sub

Public Sub ProcessSamples( _
    sngSamples() As Single, _
    ByVal lngOffset As Long, _
    ByVal lngSamples As Long _
)

    Dim result  As Single
    Dim i       As Long

    For i = lngOffset To lngOffset + lngSamples - 1
        result = fa0 * sngSamples(i) + fa1 * fX1 + fa2 * fX2 - fa3 * fY1 - fa4 * fY2

        fX2 = fX1
        fX1 = sngSamples(i)

        fY2 = fY1
        fY1 = result
        
        sngSamples(i) = result
    Next
End Sub

' various Biquad IIR Filters
' http://www.dspguru.com/sw/lib/biquad.c

' Low/High/Bandpass: dBGain not used
'                    freq is cutoff frequency
'                    srate is samplerate
'                    bandwidth in octaves
'
' Peak EQ: dBGain is boost in dB
'          freq is center frequency of the band
'          srate is samplerate
'          bandwidth in octaves
'
Public Sub CreateBiquadIIR( _
    ByVal IIRType As IIR_TYPE, _
    ByVal dBGain As Single, _
    ByVal freq As Single, _
    ByVal srate As Single, _
    ByVal bw As Single _
)

    Dim A       As Single, omega    As Single
    Dim sn      As Single, cs       As Single
    Dim Alpha   As Single, Beta     As Single

    Dim a0      As Single, a1       As Single, a2   As Single
    Dim b0      As Single, b1       As Single, b2   As Single

    lngSR = srate
    sngBW = bw
    sngDB = dBGain
    sngFreq = freq
    udeType = IIRType

    A = 10 ^ (dBGain / 40)
    omega = 2 * Pi * freq / srate
    sn = Sin(omega)
    cs = Cos(omega)
    Alpha = sn * sinh(LN2 / 2 * bw * omega / sn)
    Beta = Sqr(A + A)

    Select Case IIRType
        Case IIR_LOW_PASS
            b0 = (1 - cs) / 2
            b1 = 1 - cs
            b2 = (1 - cs) / 2
            a0 = 1 + Alpha
            a1 = -2 * cs
            a2 = 1 - Alpha
        Case IIR_HIGH_PASS
            b0 = (1 + cs) / 2
            b1 = -(1 + cs)
            b2 = (1 + cs) / 2
            a0 = 1 + Alpha
            a1 = -2 * cs
            a2 = 1 - Alpha
        Case IIR_BAND_PASS
            b0 = Alpha
            b1 = 0
            b2 = -Alpha
            a0 = 1 + Alpha
            a1 = -2 * cs
            a2 = 1 - Alpha
        Case IIR_PEAK_EQ
            b0 = 1 + (Alpha * A)
            b1 = -2 * cs
            b2 = 1 - (Alpha * A)
            a0 = 1 + (Alpha / A)
            a1 = -2 * cs
            a2 = 1 - (Alpha / A)
    End Select

    fa0 = b0 / a0
    fa1 = b1 / a0
    fa2 = b2 / a0
    fa3 = a1 / a0
    fa4 = a2 / a0

    fX1 = 0: fX2 = 0
    fY1 = 0: fY2 = 0
End Sub

Private Function sinh(angle As Single) As Single
    sinh = (Exp(angle) - Exp(-angle)) / 2#
End Function
